@@ -674,7 +674,12 @@ int array_connect(struct openconnect_info *vpninfo)
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Unexpected %d result from server\n"),
 			     ret);
-		ret = -EINVAL;
+		/* We get a redirect to /prx/000/http/localhost/cookietest when
+		 * the cookie has expired. */
+		if (ret == 302 || vpninfo->redirect_url)
+			ret = -EPERM;
+		else
+			ret = -EINVAL;
 		goto out;
 	}
 
@@ -820,6 +825,7 @@ int array_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 		int len;
 
 		if (!vpninfo->cstp_pkt) {
+			vpninfo->partial_rec_size = 0;
 			vpninfo->cstp_pkt = alloc_pkt(vpninfo, receive_mtu);
 			if (!vpninfo->cstp_pkt) {
 				vpn_progress(vpninfo, PRG_ERR, _("Allocation failed\n"));
@@ -827,45 +833,98 @@ int array_mainloop(struct openconnect_info *vpninfo, int *timeout, int readable)
 			}
 		}
 
-		len = ssl_nonblock_read(vpninfo, 0, vpninfo->cstp_pkt->data, receive_mtu);
+		len = ssl_nonblock_read(vpninfo, 0,
+					vpninfo->cstp_pkt->data + vpninfo->partial_rec_size,
+					receive_mtu - vpninfo->partial_rec_size);
 		if (!len)
 			break;
 		if (len < 0)
 			goto do_reconnect;
-		if (len < 8) {
-			vpn_progress(vpninfo, PRG_ERR, _("Short packet received (%d bytes)\n"), len);
-			vpninfo->quit_reason = "Short packet received";
-			return 1;
+
+		if (vpninfo->partial_rec_size) {
+			vpn_progress(vpninfo, PRG_DEBUG,
+				     _("Received %d more bytes after partial %d\n"),
+				     len, vpninfo->partial_rec_size);
+			len += vpninfo->partial_rec_size;
+			vpninfo->partial_rec_size = 0;
 		}
 
-		/* Check it looks like a valid IP packet, and then check for the special
-		 * IP protocol 255 that is used for control stuff. Maybe also look at length
-		 * and be prepared to *split* IP packets received in the same read() call. */
-
 		vpninfo->ssl_times.last_rx = time(NULL);
+		work_done = 1;
 
 		unsigned char *buf = vpninfo->cstp_pkt->data;
-		if (len >= sizeof(struct ip) && buf[0] == 0x45 &&
-		    load_be16(buf + 2) == len && buf[9] == 0xff) {
+		int iplen;
+	next_ip:
+		if (len < sizeof(struct ip)) {
+			vpn_progress(vpninfo, PRG_DEBUG,
+				     _("Received partial packet, %d bytes\n"), len);
+			vpninfo->partial_rec_size = len;
+			continue;
+		}
+
+		switch(buf[0] >> 4) {
+		case 4:
+			iplen = load_be16(buf + 2);
+			break;
+		case 6:
+			iplen = load_be16(buf + 4) + 40;
+			break;
+		default:
+		badiplen:
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Unrecognised data packet, len %d\n"), len);
+			dump_buf_hex(vpninfo, PRG_DEBUG, '<', buf, len);
+			continue;
+		}
+
+		if (iplen > receive_mtu)
+			goto badiplen;
+
+		/* Argh. It even splits IP packets across TLS records. */
+		if (iplen > len) {
+			vpn_progress(vpninfo, PRG_DEBUG,
+				     _("Received partial packet, %d of %d bytes\n"),
+				     len, iplen);
+			vpninfo->partial_rec_size = len;
+			continue;
+		}
+
+		/* Dump control packets but don't queue them */
+		if (buf[0] == 0x45 && buf[9] == 0xff) {
 			uint16_t ctrl_type = load_be16(buf + 12);
 			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Receive control packet of type %x:\n"),
 				     ctrl_type);
 			dump_buf_hex(vpninfo, PRG_DEBUG, '<', buf, len);
-			continue;
+			if (iplen == len)
+				continue;
+		} else {
+			vpn_progress(vpninfo, PRG_TRACE,
+				     _("Received data packet of %d bytes\n"),
+				     iplen);
+			if (iplen == len) {
+				/* This buffer (now) contains just a single IP packet */
+				vpninfo->cstp_pkt->len = iplen;
+				queue_packet(&vpninfo->incoming_queue, vpninfo->cstp_pkt);
+				vpninfo->cstp_pkt = NULL;
+				continue;
+			} else {
+				/* Awfully inefficient for now. We *copy* the packet into
+				 * a new buffer and then memmove the subsequent one(s)
+				 * down in the original buffer. */
+				queue_new_packet(vpninfo, &vpninfo->incoming_queue,
+						 buf, iplen);
+			}
 		}
 
+		/* Move the next packet(s) up to the head of the existing buffer */
+		len -= iplen;
+		memmove(buf, buf + iplen, len);
 		vpn_progress(vpninfo, PRG_TRACE,
-			     _("Received data packet of %d bytes\n"),
-			     len);
-		vpninfo->cstp_pkt->len = len;
-		queue_packet(&vpninfo->incoming_queue, vpninfo->cstp_pkt);
-		vpninfo->cstp_pkt = NULL;
-		work_done = 1;
-		continue;
+			     _("Moved down %d bytes after previous packet\n"), len);
+		goto next_ip;
 	}
 
-
 	/* If SSL_write() fails we are expected to try again. With exactly
 	   the same data, at exactly the same location. So we keep the
 	   packet we had before.... */
