@@ -99,7 +99,7 @@ int openconnect__vasprintf(char **strp, const char *fmt, va_list ap)
 		res = NULL;
 		goto err;
 	}
-	if (len >= 0 && len < 160)
+	if (len < 160)
 		goto out;
 
 	free(res);
@@ -381,11 +381,10 @@ WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 #include <afunix.h>
 #else
 #define UNIX_PATH_MAX 108
-struct sockaddr_un
-{
+struct sockaddr_un {
     ADDRESS_FAMILY sun_family;     /* AF_UNIX */
     char sun_path[UNIX_PATH_MAX];  /* pathname */
-} SOCKADDR_UN, *PSOCKADDR_UN;;
+} SOCKADDR_UN, *PSOCKADDR_UN;
 #endif /* HAS_AFUNIX_H */
 
 /* dumb_socketpair:
@@ -475,7 +474,7 @@ int dumb_socketpair(OPENCONNECT_CMD_SOCKET socks[2], int make_overlapped)
                  */
                 QueryPerformanceCounter(&ticks);
                 snprintf(a.unaddr.sun_path + n, UNIX_PATH_MAX - n,
-                         "%"PRIx64"-%"PRId32".$$$", ticks.QuadPart, GetCurrentProcessId());
+                         "%"PRIx64"-%lu.$$$", ticks.QuadPart, GetCurrentProcessId());
                 a.unaddr.sun_family = AF_UNIX;
 
                 if (bind(listener, &a.addr, addrlen) != SOCKET_ERROR)
@@ -488,7 +487,7 @@ int dumb_socketpair(OPENCONNECT_CMD_SOCKET socks[2], int make_overlapped)
 
             if (setsockopt(listener, SOL_SOCKET, SO_REUSEADDR,
                            (char *) &reuse, (socklen_t) sizeof(reuse)) == -1)
-                goto fallback;;
+                goto fallback;
 
             if (bind(listener, &a.addr, addrlen) == SOCKET_ERROR)
                 goto fallback;
