@@ -685,9 +685,9 @@ static int f5_configure(struct openconnect_info *vpninfo)
 	for (cookie = vpninfo->cookies; cookie; cookie = cookie->next) {
 		if (!strcmp(cookie->option, "F5_ST")) {
 			int junk;
-			time_t start, dur;
+			unsigned long long start, dur;
 			char c = 0;
-			if (sscanf(cookie->value, "%dz%dz%dz%ldz%ld%c", &junk, &junk, &junk, &start, &dur, &c) >= 5
+			if (sscanf(cookie->value, "%dz%dz%dz%lldz%lld%c", &junk, &junk, &junk, &start, &dur, &c) >= 5
 			    && (c == 0 || c == 'z'))
 				vpninfo->auth_expiration = start + dur;
 			break;
