@@ -18,22 +18,22 @@
 
 #include <config.h>
 
-#include <errno.h>
-#include <sys/types.h>
+#include "gnutls.h"
+
+#include <gnutls/dtls.h>
+
 #include <unistd.h>
 #include <fcntl.h>
-#include <string.h>
-#include <errno.h>
-#include <stdlib.h>
-#include <stdio.h>
+#include <sys/types.h>
 #ifndef _WIN32
 #include <netinet/in.h>
 #include <sys/socket.h>
 #endif
 
-
-#include <gnutls/dtls.h>
-#include "gnutls.h"
+#include <errno.h>
+#include <string.h>
+#include <stdlib.h>
+#include <stdio.h>
 
 #if GNUTLS_VERSION_NUMBER < 0x030400
 # define GNUTLS_CIPHER_CHACHA20_POLY1305 23
@@ -57,15 +57,15 @@ struct {
 	int cisco_dtls12;
 } gnutls_dtls_ciphers[] = {
 	{ "DHE-RSA-AES128-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_KX_DHE_RSA, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+DHE-RSA:%COMPAT", "3.0.0" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+DHE-RSA:+SIGN-ALL:%COMPAT", "3.0.0" },
 	{ "DHE-RSA-AES256-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_KX_DHE_RSA, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+DHE-RSA:%COMPAT", "3.0.0" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+DHE-RSA:+SIGN-ALL:%COMPAT", "3.0.0" },
 	{ "AES128-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT", "3.0.0" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:+SIGN-ALL:%COMPAT", "3.0.0" },
 	{ "AES256-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:%COMPAT", "3.0.0" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:+SIGN-ALL:%COMPAT", "3.0.0" },
 	{ "DES-CBC3-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_3DES_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
-	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT", "3.0.0" },
+	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:+SIGN-ALL:%COMPAT", "3.0.0" },
 	{ "OC-DTLS1_2-AES128-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_GCM, GNUTLS_KX_RSA, GNUTLS_MAC_AEAD,
 	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-GCM:+AEAD:+RSA:%COMPAT:+SIGN-ALL", "3.2.7" },
 	{ "OC-DTLS1_2-AES256-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_GCM, GNUTLS_KX_RSA, GNUTLS_MAC_AEAD,
@@ -73,6 +73,14 @@ struct {
 	{ "OC2-DTLS1_2-CHACHA20-POLY1305", GNUTLS_DTLS1_2, GNUTLS_CIPHER_CHACHA20_POLY1305, GNUTLS_KX_PSK, GNUTLS_MAC_AEAD,
 	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+CHACHA20-POLY1305:+AEAD:+PSK:%COMPAT:+SIGN-ALL", "3.4.8" },
 	/* Cisco X-DTLS12-CipherSuite: values */
+	{ "DHE-RSA-AES128-SHA", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_KX_DHE_RSA, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-CBC:+SHA1:+DHE-RSA:+SIGN-ALL:%COMPAT", "3.0.0", 1 },
+	{ "DHE-RSA-AES256-SHA", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_KX_DHE_RSA, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-CBC:+SHA1:+DHE-RSA:+SIGN-ALL:%COMPAT", "3.0.0", 1 },
+	{ "AES128-SHA", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:+SIGN-ALL:%COMPAT", "3.0.0", 1 },
+	{ "AES256-SHA", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_KX_RSA, GNUTLS_MAC_SHA1,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:+SIGN-ALL:%COMPAT", "3.0.0", 1 },
 	{ "ECDHE-RSA-AES256-GCM-SHA384", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_GCM, GNUTLS_KX_ECDHE_RSA, GNUTLS_MAC_AEAD,
 	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-GCM:+AEAD:+ECDHE-RSA:+SIGN-ALL:%COMPAT", "3.2.7", 1 },
 	{ "ECDHE-RSA-AES128-GCM-SHA256", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_GCM, GNUTLS_KX_ECDHE_RSA, GNUTLS_MAC_AEAD,
@@ -94,7 +102,7 @@ void gather_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 {
 	int i, first = 1;
 
-	for (i = 0; i < sizeof(gnutls_dtls_ciphers) / sizeof(gnutls_dtls_ciphers[0]); i++) {
+	for (i = 0; i < ARRAY_SIZE(gnutls_dtls_ciphers); i++) {
 		if (!gnutls_dtls_ciphers[i].cisco_dtls12 &&
 		    gnutls_check_version(gnutls_dtls_ciphers[i].min_gnutls_version)) {
 			buf_append(buf, "%s%s", first ? "" : ":",
@@ -132,7 +140,7 @@ void gather_dtls_ciphers(struct openconnect_info *vpninfo, struct oc_text_buf *b
 			break;
 
 		if (gnutls_cipher_suite_info(idx, NULL, NULL, &cipher, &mac, NULL) != NULL) {
-			for (i = 0; i < sizeof(gnutls_dtls_ciphers)/sizeof(gnutls_dtls_ciphers[0]); i++) {
+			for (i = 0; i < ARRAY_SIZE(gnutls_dtls_ciphers); i++) {
 				if (used & (1 << i))
 					continue;
 				if (gnutls_dtls_ciphers[i].mac == mac && gnutls_dtls_ciphers[i].cipher == cipher) {
@@ -230,7 +238,7 @@ static int start_dtls_psk_handshake(struct openconnect_info *vpninfo, gnutls_ses
 	 * the RFC5705 context, the output is identical with gnutls_prf(). The
 	 * latter is available in much earlier versions of gnutls. */
 	err = gnutls_prf(vpninfo->https_sess, PSK_LABEL_SIZE, PSK_LABEL,
-			 0, 0, 0, PSK_KEY_SIZE, (char*)vpninfo->dtls_secret);
+			 0, 0, 0, PSK_KEY_SIZE, (char *)vpninfo->dtls_secret);
 	if (err < 0) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to generate DTLS key: %s\n"),
@@ -276,8 +284,8 @@ static int start_dtls_resume_handshake(struct openconnect_info *vpninfo, gnutls_
 	int err;
 	int cipher;
 
-	for (cipher = 0; cipher < sizeof(gnutls_dtls_ciphers)/sizeof(gnutls_dtls_ciphers[0]); cipher++) {
-		if (gnutls_dtls_ciphers[cipher].cisco_dtls12 != vpninfo->cisco_dtls12 ||
+	for (cipher = 0; cipher < ARRAY_SIZE(gnutls_dtls_ciphers); cipher++) {
+		if (gnutls_dtls_ciphers[cipher].cisco_dtls12 != vpninfo->dtls12 ||
 		    gnutls_check_version(gnutls_dtls_ciphers[cipher].min_gnutls_version) == NULL)
 			continue;
 		if (!strcmp(vpninfo->dtls_cipher, gnutls_dtls_ciphers[cipher].name))
@@ -317,6 +325,44 @@ static int start_dtls_resume_handshake(struct openconnect_info *vpninfo, gnutls_
 	return 0;
 }
 
+static int start_dtls_anon_handshake(struct openconnect_info *vpninfo, gnutls_session_t dtls_ssl)
+{
+	char *prio = vpninfo->ciphersuite_config;
+	int ret;
+
+	/*
+	 * Use the same cred store as for the HTTPS session. That has
+	 * our own verify_peer() callback installed, and will validate
+	 * just like we do for the HTTPS service.
+	 *
+	 * There is also perhaps a case to be made for *only* accepting
+	 * precisely the same cert that we get from the HTTPS service,
+	 * but we tried that for EAP-TTLS in the Pulse protocol and the
+	 * theory was disproven, so we ended up doing this there too.
+         */
+	gnutls_credentials_set(dtls_ssl, GNUTLS_CRD_CERTIFICATE, vpninfo->https_cred);
+
+	/* The F5 BIG-IP server before v16, will crap itself if we
+	 * even *try* to do DTLSv1.2 */
+	if (!vpninfo->dtls12 &&
+	    asprintf(&prio, "%s:-VERS-DTLS1.2:+VERS-DTLS1.0",
+		     vpninfo->ciphersuite_config) < 0)
+		return -ENOMEM;
+
+	ret = gnutls_priority_set_direct(dtls_ssl, prio ? : vpninfo->ciphersuite_config,
+					 NULL);
+	if (ret) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to set DTLS priority: '%s': %s\n"),
+			     prio, gnutls_strerror(ret));
+	}
+
+	if (prio != vpninfo->ciphersuite_config)
+		free(prio);
+
+	return ret;
+}
+
 /*
  * GnuTLS version between 3.6.3 and 3.6.12 send zero'ed ClientHello. Make sure
  * we are not hitting that bug. Adapted from:
@@ -375,10 +421,16 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	gnutls_transport_set_ptr(dtls_ssl,
 				 (gnutls_transport_ptr_t)(intptr_t)dtls_fd);
 
-	if (!strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE"))
+	if (!vpninfo->dtls_cipher) {
+		/* Anonymous DTLS (PPP protocols) */
+		ret = start_dtls_anon_handshake(vpninfo, dtls_ssl);
+	} else if (!strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE")) {
+		/* For OpenConnect/ocserv protocol */
 		ret = start_dtls_psk_handshake(vpninfo, dtls_ssl);
-	else
+	} else {
+		/* Nonexistent session resume hack (Cisco AnyConnect) */
 		ret = start_dtls_resume_handshake(vpninfo, dtls_ssl);
+	}
 
 	if (ret) {
 		if (ret != -EAGAIN)
@@ -397,14 +449,29 @@ int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 	return 0;
 }
 
-int dtls_try_handshake(struct openconnect_info *vpninfo)
+int dtls_try_handshake(struct openconnect_info *vpninfo, int *timeout)
 {
 	int err = gnutls_handshake(vpninfo->dtls_ssl);
 	char *str;
 
 	if (!err) {
-		if (!strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE")) {
-			/* For PSK-NEGOTIATE, we have to determine the tunnel MTU
+		if (!vpninfo->dtls_cipher) {
+			/* Anonymous DTLS (PPP protocols) will set vpninfo->ip_info.mtu
+			 * in PPP negotiation.
+			 *
+			 * XX: Needs forthcoming overhaul to detect MTU correctly and offer
+			 * reasonable MRU values during PPP negotiation.
+			 */
+			int data_mtu = vpninfo->cstp_basemtu = 1500;
+			if (vpninfo->peer_addr->sa_family == IPPROTO_IPV6)
+				data_mtu -= 40; /* IPv6 header */
+			else
+				data_mtu -= 20; /* Legacy IP header */
+			data_mtu -= 8; /* UDP header */
+			dtls_set_mtu(vpninfo, data_mtu);
+		} else if (!strcmp(vpninfo->dtls_cipher, "PSK-NEGOTIATE")) {
+			/* For PSK-NEGOTIATE (OpenConnect/ocserv protocol)
+			 * we have to determine the tunnel MTU
 			 * for ourselves based on the base MTU */
 			int data_mtu = vpninfo->cstp_basemtu;
 			if (vpninfo->peer_addr->sa_family == IPPROTO_IPV6)
@@ -428,6 +495,7 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 				vpninfo->ip_info.mtu = data_mtu;
 			}
 		} else {
+			/* Nonexistent session resume hack (Cisco AnyConnect) */
 			if (!gnutls_session_is_resumed(vpninfo->dtls_ssl)) {
 				/* Someone attempting to hijack the DTLS session?
 				 * A real server would never allow a full session
@@ -467,7 +535,7 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 			}
 		}
 
-		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx = 
+		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx =
 			vpninfo->dtls_times.last_tx = time(NULL);
 
 		dtls_detect_mtu(vpninfo);
@@ -476,8 +544,18 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 	}
 
 	if (err == GNUTLS_E_AGAIN || err == GNUTLS_E_INTERRUPTED) {
-		if (time(NULL) < vpninfo->new_dtls_started + 12)
+		int quit_time = vpninfo->new_dtls_started + 12 - time(NULL);
+		if (quit_time > 0) {
+			if (timeout) {
+				unsigned next_resend = gnutls_dtls_get_timeout(vpninfo->dtls_ssl);
+				if (next_resend && *timeout > next_resend)
+					*timeout = next_resend;
+
+				if (*timeout > quit_time * 1000)
+					*timeout = quit_time * 1000;
+			}
 			return 0;
+		}
 		vpn_progress(vpninfo, PRG_DEBUG, _("DTLS handshake timed out\n"));
 	}
 
@@ -491,6 +569,8 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 
 	vpninfo->dtls_state = DTLS_SLEEPING;
 	time(&vpninfo->new_dtls_started);
+	if (timeout && *timeout > vpninfo->dtls_attempt_period * 1000)
+		*timeout = vpninfo->dtls_attempt_period * 1000;
 	return -EINVAL;
 }
 
