@@ -132,6 +132,7 @@ public abstract class LibOpenConnect {
 	public synchronized native void setPassTOS(boolean isEnabled);
 	public synchronized native int setProxyAuth(String methods);
 	public synchronized native int setHTTPProxy(String proxy);
+	public synchronized native void setSNI(String sni);
 	public synchronized native void setXMLSHA1(String hash);
 	public synchronized native void setHostname(String hostname);
 	public synchronized native void setVersionString(String version);
@@ -270,7 +271,6 @@ public abstract class LibOpenConnect {
 		public String proxyPac;
 		public String gatewayAddr;
 		public int MTU;
-		public int idleTimeoutSec;
 
 		public ArrayList<String> splitDNS = new ArrayList<String>();
 		public ArrayList<String> splitIncludes = new ArrayList<String>();
