@@ -41,6 +41,8 @@
 #include LIBPROXY_HDR
 #endif
 
+#define MAX_READ_STDIN_SIZE 4096
+
 #ifdef _WIN32
 #include <shlwapi.h>
 #include <wtypes.h>
@@ -96,8 +98,6 @@ static int allow_stdin_read;
 static char *token_filename;
 static int allowed_fingerprints;
 
-static char *ext_browser;
-
 struct accepted_cert {
 	struct accepted_cert *next;
 	char *fingerprint;
@@ -438,7 +438,7 @@ static void read_stdin(char **string, int hidden, int allow_fail)
 	CONSOLE_READCONSOLE_CONTROL rcc = { sizeof(rcc), 0, 13, 0 };
 	HANDLE stdinh = GetStdHandle(STD_INPUT_HANDLE);
 	DWORD cmode, nr_read, last_error;
-	wchar_t wbuf[1024];
+	wchar_t wbuf[MAX_READ_STDIN_SIZE];
 	char *buf;
 
 	if (GetConsoleMode(stdinh, &cmode)) {
@@ -767,7 +767,7 @@ static void print_supported_protocols_usage(void)
 static const char default_vpncscript[] = DEFAULT_VPNCSCRIPT;
 static void read_stdin(char **string, int hidden, int allow_fail)
 {
-	char *c, *got, *buf = malloc(1025);
+	char *c, *got, *buf = malloc(MAX_READ_STDIN_SIZE+1);
 	int fd = fileno(stdin);
 	struct termios t;
 
@@ -782,7 +782,7 @@ static void read_stdin(char **string, int hidden, int allow_fail)
 		tcsetattr(fd, TCSANOW, &t);
 	}
 
-	got = fgets(buf, 1025, stdin);
+	got = fgets(buf, MAX_READ_STDIN_SIZE+1, stdin);
 
 	if (hidden) {
 		t.c_lflag |= ECHO;
@@ -902,34 +902,6 @@ static BOOL WINAPI console_ctrl_handler(DWORD dwCtrlType)
 }
 #endif
 
-#ifdef HAVE_POSIX_SPAWN
-static int spawn_browser(struct openconnect_info *vpninfo, const char *url, void *cbdata)
-{
-	vpn_progress(vpninfo, PRG_TRACE,
-		     _("Main Spawning external browser '%s'\n"),
-		     ext_browser);
-
-	pid_t pid = 0;
-	char *browser_argv[3] = { ext_browser, (char *)url, NULL };
-	posix_spawn_file_actions_t file_actions, *factp = NULL;
-	int ret = 0;
-
-	if (!posix_spawn_file_actions_init(&file_actions)) {
-		factp = &file_actions;
-
-		posix_spawn_file_actions_adddup2(&file_actions, STDERR_FILENO, STDOUT_FILENO);
-	}
-
-	if (posix_spawn(&pid, ext_browser, factp, NULL, browser_argv, environ)) {
-		ret = -errno;
-		vpn_perror(vpninfo, _("Spawn browser"));
-	}
-
-	if (factp)
-		posix_spawn_file_actions_destroy(factp);
-	return ret;
-}
-#endif
 static void print_default_vpncscript(void)
 {
 	printf("%s %s\n", _("Default vpnc-script (override with --script):"),
@@ -2089,7 +2061,7 @@ int main(int argc, char **argv)
 			vpnc_script = dup_config_arg();
 			break;
 		case OPT_EXT_BROWSER:
-			ext_browser = dup_config_arg();
+			vpninfo->external_browser = dup_config_arg();
 			break;
 		case OPT_NO_EXTERNAL_AUTH:
 			/* XX: Is this a workaround for a server bug, or a "normal" authentication option? */
@@ -2280,11 +2252,6 @@ int main(int argc, char **argv)
 	if (proxy && openconnect_set_http_proxy(vpninfo, strdup(proxy)))
 		exit(1);
 
-#ifdef HAVE_POSIX_SPAWN
-	if (ext_browser)
-		openconnect_set_external_browser_callback(vpninfo, spawn_browser);
-#endif
-
 #ifndef _WIN32
 	memset(&sa, 0, sizeof(sa));
 
@@ -3005,11 +2972,8 @@ static void init_token(struct openconnect_info *vpninfo,
 		case -EINVAL:
 			fprintf(stderr, _("Soft token string is invalid\n"));
 			exit(1);
-		case -EOPNOTSUPP:
-			fprintf(stderr, _("OpenConnect was not built with liboath support\n"));
-			exit(1);
 		default:
-			fprintf(stderr, _("General failure in liboath\n"));
+			fprintf(stderr, _("General failure in TOTP/HOTP support\n"));
 			exit(1);
 		}
 
