@@ -94,7 +94,10 @@
 #ifndef _Ret_bytecount_
 #define _Ret_bytecount_(sz)
 #endif
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wunknown-pragmas"
 #include "wintun.h"
+#pragma GCC diagnostic pop
 
 #include <ws2tcpip.h>
 #ifndef SECURITY_WIN32
@@ -122,8 +125,11 @@
 #if defined(__APPLE__)
 #include <crt_externs.h>
 #define environ (*_NSGetEnviron())
-#elif defined(__sun__)
-// Is there nothing we can #include for this? And no consts?
+#else
+/*
+ * POSIX.1-2017 says that environ must be declared by the user if it is to be used directly:
+ * https://pubs.opengroup.org/onlinepubs/9699919799/functions/exec.html
+ */
 extern char **environ;
 #endif
 #include <spawn.h>
@@ -475,6 +481,7 @@ struct openconnect_info {
 	struct http_auth_state http_auth[MAX_AUTH_TYPES];
 	struct http_auth_state proxy_auth[MAX_AUTH_TYPES];
 	int no_external_auth;
+	const char *external_browser;
 
 	char *localname;
 
