@@ -38,7 +38,7 @@
 #include <string.h>
 #include <ctype.h>
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
 #define X509_up_ref(x) CRYPTO_add(&(x)->references, 1, CRYPTO_LOCK_X509)
 #define X509_get0_notAfter(x) X509_get_notAfter(x)
 #define EVP_MD_CTX_new EVP_MD_CTX_create
@@ -49,6 +49,8 @@
 typedef int (*X509_STORE_CTX_get_issuer_fn)(X509 **issuer,
 					    X509_STORE_CTX *ctx, X509 *x);
 #define X509_STORE_CTX_get_get_issuer(ctx) ((ctx)->get_issuer)
+#define OpenSSL_version SSLeay_version
+#define OPENSSL_VERSION SSLEAY_VERSION
 #endif
 
 static char tls_library_version[32] = "";
@@ -56,7 +58,9 @@ static char tls_library_version[32] = "";
 const char *openconnect_get_tls_library_version(void)
 {
 	if (!*tls_library_version) {
-		strncpy(tls_library_version, SSLeay_version(SSLEAY_VERSION), sizeof(tls_library_version));
+		strncpy(tls_library_version,
+			OpenSSL_version(OPENSSL_VERSION),
+			sizeof(tls_library_version));
 		tls_library_version[sizeof(tls_library_version)-1]='\0';
 	}
 	return tls_library_version;
@@ -170,7 +174,13 @@ static int _openconnect_openssl_write(SSL *ssl, int fd, struct openconnect_info
 				return -EIO;
 			}
 			cmd_fd_set(vpninfo, &rd_set, &maxfd);
-			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+			while (select(maxfd + 1, &rd_set, &wr_set, NULL, NULL) < 0) {
+				if (errno != EINTR) {
+					vpn_perror(vpninfo, _("Failed select() for TLS/DTLS"));
+					return -EIO;
+				}
+			}
+
 			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("TLS/DTLS write cancelled\n"));
 				return -EINTR;
@@ -220,7 +230,12 @@ static int _openconnect_openssl_read(SSL *ssl, int fd, struct openconnect_info *
 			return -EIO;
 		}
 		cmd_fd_set(vpninfo, &rd_set, &maxfd);
-		ret = select(maxfd + 1, &rd_set, &wr_set, NULL, tv);
+		while ((ret = select(maxfd + 1, &rd_set, &wr_set, NULL, tv)) < 0) {
+			if (errno != EINTR) {
+				vpn_perror(vpninfo, _("Failed select() for TLS/DTLS"));
+				return -EIO;
+			}
+		}
 		if (is_cancel_pending(vpninfo, &rd_set)) {
 			vpn_progress(vpninfo, PRG_ERR, _("TLS/DTLS read cancelled\n"));
 			return -EINTR;
@@ -287,7 +302,12 @@ static int openconnect_openssl_gets(struct openconnect_info *vpninfo, char *buf,
 				break;
 			}
 			cmd_fd_set(vpninfo, &rd_set, &maxfd);
-			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+			while (select(maxfd + 1, &rd_set, &wr_set, NULL, NULL) < 0) {
+				if (errno != EINTR) {
+					vpn_perror(vpninfo, _("Failed select() for TLS/DTLS"));
+					return -EIO;
+				}
+			}
 			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("TLS/DTLS read cancelled\n"));
 				ret = -EINTR;
@@ -1205,7 +1225,7 @@ static int set_peer_cert_hash(struct openconnect_info *vpninfo)
 	return 0;
 }
 
-#if OPENSSL_VERSION_NUMBER < 0x10002000L || defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER < 0x10002000L
 static int match_hostname_elem(const char *hostname, int helem_len,
 			       const char *match, int melem_len)
 {
@@ -1274,8 +1294,8 @@ static int match_hostname(const char *hostname, const char *match)
 }
 
 /* cf. RFC2818 and RFC2459 */
-static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert,
-			       const unsigned char *ipaddr, int ipaddrlen)
+static int match_cert_hostname_or_ip(struct openconnect_info *vpninfo, X509 *peer_cert,
+				     char *hostname)
 {
 	STACK_OF(GENERAL_NAME) *altnames;
 	X509_NAME *subjname;
@@ -1284,6 +1304,21 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 	int i, altdns = 0;
 	int ret;
 
+	unsigned char ipaddr[sizeof(struct in6_addr)];
+	int ipaddrlen = 0;
+	if (inet_pton(AF_INET, hostname, ipaddr) > 0)
+		ipaddrlen = 4;
+	else if (inet_pton(AF_INET6, hostname, ipaddr) > 0)
+		ipaddrlen = 16;
+	else if (hostname[0] == '[' &&
+		 hostname[strlen(hostname)-1] == ']') {
+		char *p = &hostname[strlen(hostname)-1];
+		*p = 0;
+		if (inet_pton(AF_INET6, hostname + 1, ipaddr) > 0)
+			ipaddrlen = 16;
+		*p = ']';
+	}
+
 	altnames = X509_get_ext_d2i(peer_cert, NID_subject_alt_name,
 				    NULL, NULL);
 	for (i = 0; i < sk_GENERAL_NAME_num(altnames); i++) {
@@ -1302,7 +1337,7 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 			if (strlen(str) != len)
 				continue;
 
-			if (!match_hostname(vpninfo->hostname, str)) {
+			if (!match_hostname(hostname, str)) {
 				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("Matched DNS altname '%s'\n"),
 					     str);
@@ -1373,14 +1408,14 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 
 			/* Leave url_host as it was so that it can be freed */
 			url_host2 = url_host;
-			if (ipaddrlen == 16 && vpninfo->hostname[0] != '[' &&
+			if (ipaddrlen == 16 && hostname[0] != '[' &&
 			    url_host[0] == '[' && url_host[strlen(url_host)-1] == ']') {
 				/* Cope with https://[IPv6]/ when the hostname is bare IPv6 */
 				url_host[strlen(url_host)-1] = 0;
 				url_host2++;
 			}
 
-			if (strcasecmp(vpninfo->hostname, url_host2))
+			if (strcasecmp(hostname, url_host2))
 				goto no_uri_match;
 
 			if (url_path) {
@@ -1417,7 +1452,7 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 	if (altdns) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("No altname in peer cert matched '%s'\n"),
-			     vpninfo->hostname);
+			     hostname);
 		return -EINVAL;
 	}
 
@@ -1464,11 +1499,26 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 	return ret;
 }
 #else
-static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert,
-			       const unsigned char *ipaddr, int ipaddrlen)
+static int match_cert_hostname_or_ip(struct openconnect_info *vpninfo, X509 *peer_cert,
+				     char *hostname)
 {
 	char *matched = NULL;
 
+	unsigned char ipaddr[sizeof(struct in6_addr)];
+	int ipaddrlen = 0;
+	if (inet_pton(AF_INET, hostname, ipaddr) > 0)
+		ipaddrlen = 4;
+	else if (inet_pton(AF_INET6, hostname, ipaddr) > 0)
+		ipaddrlen = 16;
+	else if (hostname[0] == '[' &&
+		 hostname[strlen(hostname)-1] == ']') {
+		char *p = &hostname[strlen(hostname)-1];
+		*p = 0;
+		if (inet_pton(AF_INET6, hostname + 1, ipaddr) > 0)
+			ipaddrlen = 16;
+		*p = ']';
+	}
+
 	if (ipaddrlen && X509_check_ip(peer_cert, ipaddr, ipaddrlen, 0) == 1) {
 		if (vpninfo->verbose >= PRG_DEBUG) {
 			char host[80];
@@ -1488,7 +1538,7 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 		}
 		return 0;
 	}
-	if (X509_check_host(peer_cert, vpninfo->hostname, 0, 0, &matched) == 1) {
+	if (X509_check_host(peer_cert, hostname, 0, 0, &matched) == 1) {
 		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Matched peer certificate subject name '%s'\n"),
 			     matched);
@@ -1639,23 +1689,12 @@ static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 	if (!X509_verify_cert(ctx)) {
 		err_string = X509_verify_cert_error_string(X509_STORE_CTX_get_error(ctx));
 	} else {
-		unsigned char addrbuf[sizeof(struct in6_addr)];
-		int addrlen = 0;
-
-		if (inet_pton(AF_INET, vpninfo->hostname, addrbuf) > 0)
-			addrlen = 4;
-		else if (inet_pton(AF_INET6, vpninfo->hostname, addrbuf) > 0)
-			addrlen = 16;
-		else if (vpninfo->hostname[0] == '[' &&
-			 vpninfo->hostname[strlen(vpninfo->hostname)-1] == ']') {
-			char *p = &vpninfo->hostname[strlen(vpninfo->hostname)-1];
-			*p = 0;
-			if (inet_pton(AF_INET6, vpninfo->hostname + 1, addrbuf) > 0)
-				addrlen = 16;
-			*p = ']';
-		}
-
-		if (match_cert_hostname(vpninfo, vpninfo->peer_cert, addrbuf, addrlen))
+		if (vpninfo->sni && vpninfo->sni[0]) {
+			if (match_cert_hostname_or_ip(vpninfo, vpninfo->peer_cert, vpninfo->sni))
+				err_string = _("certificate does not match SNI");
+			else
+				return 1;
+		} else if (match_cert_hostname_or_ip(vpninfo, vpninfo->peer_cert, vpninfo->hostname))
 			err_string = _("certificate does not match hostname");
 		else
 			return 1;
@@ -1682,7 +1721,7 @@ static int ssl_app_verify_callback(X509_STORE_CTX *ctx, void *arg)
 static int check_certificate_expiry(struct openconnect_info *vpninfo, struct cert_info *certinfo,
 				    struct ossl_cert_info *oci)
 {
-	method_const ASN1_TIME *notAfter;
+	const ASN1_TIME *notAfter;
 	const char *reason = NULL;
 	time_t t;
 	int i;
@@ -1839,7 +1878,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 		return ssl_sock;
 
 	if (!vpninfo->https_ctx) {
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
 		vpninfo->https_ctx = SSL_CTX_new(SSLv23_client_method());
 		if (vpninfo->https_ctx)
 			SSL_CTX_set_options(vpninfo->https_ctx, SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3);
@@ -1963,8 +2002,10 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	 * 4fcdd66fff5fea0cfa1055c6680a76a4303f28a2
 	 * cd6bd5ffda616822b52104fee0c4c7d623fd4f53
 	 */
-#if OPENSSL_VERSION_NUMBER >= 0x10001070 && !defined(LIBRESSL_VERSION_NUMBER)
-	if (string_is_hostname(vpninfo->hostname))
+#if OPENSSL_VERSION_NUMBER >= 0x10001070L
+	if (vpninfo->sni)
+		SSL_set_tlsext_host_name(https_ssl, vpninfo->sni);
+	else if (string_is_hostname(vpninfo->hostname))
 		SSL_set_tlsext_host_name(https_ssl, vpninfo->hostname);
 #endif
 	SSL_set_verify(https_ssl, SSL_VERIFY_PEER, NULL);
@@ -2058,7 +2099,7 @@ int openconnect_init_ssl(void)
 	if (ret)
 		return ret;
 #endif
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || defined(LIBRESSL_VERSION_NUMBER)
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
 	SSL_library_init();
 	ERR_clear_error();
 	SSL_load_error_strings();
@@ -2423,7 +2464,7 @@ int ecdh_compute_secp256r1(struct openconnect_info *vpninfo, const unsigned char
 }
 
 int hkdf_sha256_extract_expand(struct openconnect_info *vpninfo, unsigned char *buf,
-			       const char *info, int infolen)
+			       const unsigned char *info, int infolen)
 {
 	size_t buflen = 32;
 	int ret = 0;
